#define OLC_PGE_APPLICATION
#include "olcPixelGameEngine.h"
#define OLC_PGEX_QUICKGUI
#include "olcPGEX_QuickGUI.h"

using namespace olc;

#define PI 3.14159

// Override base class with your custom functionality
class CardGameRef : public olc::PixelGameEngine
{
	enum class Color{
		RED,
		GREEN,
		BLUE,
		YELLOW
	};

	enum CardValue{
		ZERO,
		ONE,
		TWO,
		THREE,
		FOUR,
		FIVE,
		SIX,
		SEVEN,
		EIGHT,
		NINE,
		SKIP,
		REVERSE,
		DRAWTWO,
		WILD,
		DRAWFOUR,
	};

	enum GameState{
		AIPICK,
		DRAW,
		PICK,
		RESULT,
		PICKCOLOR,
		VICTORY
	};

	struct Rect{
		vf2d pos;
		vf2d size;
	};

	struct Card{
		Color col;
		CardValue val;
		bool faceDown=false;
		Card(Color col,CardValue val,bool faceDown=false)
			:col(col),val(val),faceDown(faceDown){}
	};

	typedef std::vector<Card> Hand;
	typedef std::vector<Card> CardPile;

	std::vector<Decal*>IMAGES;
	Rect cardBorder,cardBack,skip,reverse,wild,drawTwo,drawFour;
	Decal*cardTilesheet;
	std::vector<Hand>hands;
	CardPile startingPile,drawPile,discardPile;
	QuickGUI::Manager playerPickMenu;
	QuickGUI::Button *TwoPlayerButton,*ThreePlayerButton,*FourPlayerButton;
	vf2d drawPilePos={100,88};
	vf2d discardPilePos={156,88};

	GameState state=AIPICK;

	int playerTurn=0;

public:
	CardGameRef()
	{
		// Name your application
		sAppName = "Uno!";
	}

	bool TurnRequiresDraw(){
		Card&topCard=discardPile.back();
		for(Card&card:hands[playerTurn]){
			if(card.col==topCard.col || card.val==topCard.val || card.val==WILD||card.val==DRAWFOUR){
				return false;
			}
		}
		return true;
	}

	void ReshuffleDrawPile(){
		while(discardPile.size()>1){
			int randomCardIndex=rand()%(discardPile.size()-1);
			drawPile.push_back(discardPile[randomCardIndex]);
			drawPile.back().faceDown=true;
			discardPile.erase(discardPile.begin()+randomCardIndex);
		}
	}

	void DrawCard(Hand&hand,bool faceUp=false){
		if(drawPile.size()==0){
			ReshuffleDrawPile();
		}
		if(drawPile.size()>0){//It's possible all the cards in the game may have been drawn, 
			hand.push_back(drawPile.back());
			hand.back().faceDown=!faceUp;
			drawPile.erase(drawPile.end()-1);
		}
	}

	void DisplayCardPile(vf2d pos,CardPile&pile){
		if(pile.size()>0){
			for(int i=0;i<pile.size()/10+1;i++){
				DrawRotatedDecal(pos+vf2d{float(i),-float(i)}/2,GetCardImage(Color::RED,CardValue::ZERO,true),0,{24,32});
			}
			DrawRotatedDecal(pos+vf2d{float(pile.size()/10),-float(pile.size()/10)}/2,GetCardImage(pile.back().col,pile.back().val,pile.back().faceDown),0,{24,32});
		}
	}

	float degToRad(float deg){
		return deg*(PI/180);
	}

	float radToDeg(float rad){
		return rad*57.2957795130823208767;
	}

	void DrawGameBoard(){
		DisplayCardPile(drawPilePos,drawPile);
		DisplayCardPile(discardPilePos,discardPile);

		vf2d playerHandCenter = {128,220};
		float cursorDir = radToDeg(atan2(playerHandCenter.y-GetMouseY(),playerHandCenter.x-GetMouseX()))-180;
		float cursorDist = sqrt(pow(playerHandCenter.x-GetMouseX(),2)+pow(playerHandCenter.y-GetMouseY(),2));
		Hand&playerHand=hands[0];
		float spreadAngle=15;
		float currentAngle=-90;
		float cardDistance=64;
		if(playerHand.size()>8){ //Handle a case where the hand size is really large. We max out at 120 degrees.
			spreadAngle=120.f/playerHand.size();
			currentAngle=-60+spreadAngle/2-90;
		} else { //Small hand size, we can slowly fan out further.
			currentAngle=-float(playerHand.size()-1)/2*15-90;
			spreadAngle=15;
		}

		struct CardDrawData{
			vf2d drawPos;
			float angle;
			Card*card=nullptr;
		};

		CardDrawData selectedCardData;
		int cardIndex=0;
		for(int i=0;i<playerHand.size();i++){
			Card&currentCard=playerHand[cardIndex];
			if(cursorDist>32&&cursorDist<100&&abs(cursorDir-currentAngle)<spreadAngle/2){
				selectedCardData={playerHandCenter+vf2d{cos(degToRad(currentAngle)),sin(degToRad(currentAngle))}*(cardDistance+10),degToRad(currentAngle+90),&currentCard};
			} else {
				DrawRotatedDecal(playerHandCenter+vf2d{cos(degToRad(currentAngle)),sin(degToRad(currentAngle))}*cardDistance,GetCardImage(currentCard.col,currentCard.val,currentCard.faceDown),degToRad(currentAngle+90),{24,32},{0.6,0.6},{128,128,128,255});
			}
			currentAngle+=spreadAngle;
			cardIndex++;
		}
		if(selectedCardData.card!=nullptr){
			Card&card=*selectedCardData.card;
			DrawRotatedDecal(selectedCardData.drawPos,GetCardImage(card.col,card.val,card.faceDown),selectedCardData.angle,{24,32},{0.6,0.6});
		}
	}

	void GenerateCards(){
		for(int color=0;color<4;color++){
			for(int numb=0;numb<13;numb++){
				Decal*card=new Decal(new Sprite(48,64));
				SetDrawTarget(card->sprite);
				DrawPartialSprite({0,0},cardTilesheet->sprite,cardBack.pos,cardBack.size);
				Pixel targetCol;
				switch(Color(color)){
				case Color::RED:{
					targetCol={192,0,0};
				}break;
				case Color::GREEN:{
					targetCol={0,192,0};
				}break;
				case Color::BLUE:{
					targetCol={0,0,192};
				}break;
				case Color::YELLOW:{
					targetCol={192,192,0};
				}break;
				}
				for(int y=0;y<64;y++){
					for(int x=0;x<48;x++){
						if(card->sprite->GetPixel(x,y)==Pixel{235,235,235}){
							card->sprite->SetPixel({x,y},targetCol);
						} else 
						if(card->sprite->GetPixel(x,y)==Pixel{103,103,103}){
							card->sprite->SetPixel({x,y},targetCol/2);
						} else 
						if(card->sprite->GetPixel(x,y)==Pixel{0,0,0}){
							card->sprite->SetPixel({x,y},{0,0,0,0});
						}
					}
				}
				DrawPartialSprite({0,0},cardTilesheet->sprite,cardBorder.pos,cardBorder.size);
				switch(numb){
					case SKIP:{
						DrawPartialSprite(vi2d{24,32}-skip.size/2,cardTilesheet->sprite,skip.pos,skip.size);
					}break;
					case REVERSE:{
						DrawPartialSprite(vi2d{24,32}-reverse.size/2,cardTilesheet->sprite,reverse.pos,reverse.size);
					}break;
					case DRAWTWO:{
						DrawPartialSprite(vi2d{24,32}-drawTwo.size/2,cardTilesheet->sprite,drawTwo.pos,drawTwo.size);
					}break;
					default:{
						DrawString(vi2d{24,32}-GetTextSize(std::to_string(numb))/2*3,std::to_string(numb),WHITE,3);
					}
				}
				card->Update();
				IMAGES.push_back(card);
			}
		}
		for(int i=0;i<3;i++){
			Decal*card=new Decal(new Sprite(48,64));
			SetDrawTarget(card->sprite);
			DrawPartialSprite({0,0},cardTilesheet->sprite,cardBack.pos,cardBack.size);
			switch(i){
			case 0:{//Wild Card
				DrawPartialSprite(vi2d{24,32}-wild.size/2,cardTilesheet->sprite,wild.pos,wild.size);
			}break;
			case 1:{//Draw Four Card
				DrawPartialSprite(vi2d{24,32}-drawFour.size/2,cardTilesheet->sprite,drawFour.pos,drawFour.size);
			}break;
			}
			switch(i){//Two cards have black backs.
			case 0: //Wild Card
			case 1:{//Draw Four Card
				for(int y=0;y<64;y++){
					for(int x=0;x<48;x++){
						if(card->sprite->GetPixel(x,y)==Pixel{235,235,235}){
							card->sprite->SetPixel({x,y},BLACK);
						} else 
							if(card->sprite->GetPixel(x,y)==Pixel{103,103,103}){
								card->sprite->SetPixel({x,y},BLACK/2);
							}
					}
				}
			}break;
			default:{//This is the face down card.
				for(int y=0;y<64;y++){
					for(int x=0;x<48;x++){
						if(card->sprite->GetPixel(x,y)==Pixel{235,235,235}){
							card->sprite->SetPixel({x,y},BLACK);
						} else 
							if(card->sprite->GetPixel(x,y)==Pixel{103,103,103}){
								card->sprite->SetPixel({x,y},{102,0,64});
							}
					}
				}
			}
			}
			DrawPartialSprite({0,0},cardTilesheet->sprite,cardBorder.pos,cardBorder.size);
			card->Update();
			IMAGES.push_back(card);
		}
		SetDrawTarget(nullptr);
	}

public:
	bool OnUserCreate() override
	{
		srand(time(NULL));
		// Called once at the start, so create things here

		cardBorder={{0,0},{48,64}};
		cardBack={{48,0},{48,64}};
		skip={{0,64},{32,32}};
		reverse={{32,64},{32,32}};
		wild={{64,64},{32,32}};
		drawTwo={{96,32},{32,32}};
		drawFour={{96,64},{32,32}};

		cardTilesheet = new Decal(new Sprite("CardGame.png"));

		SetPixelMode(Pixel::MASK);

		GenerateCards();

		//2 of each basic card. (80)
		//2 of each special card. (24)
		//8 Wild Cards (8)
		//4 Draw Four Cards (4)
		for(int color=0;color<4;color++){
			for(int value=0;value<13;value++){
				startingPile.emplace_back(Color(color),CardValue(value),true);
				startingPile.emplace_back(Color(color),CardValue(value),true);
			}
		}
		for(int i=0;i<8;i++){
			startingPile.emplace_back(Color::RED,WILD,true);
		}
		for(int i=0;i<4;i++){
			startingPile.emplace_back(Color::RED,DRAWFOUR,true);
		}

		while(startingPile.size()>0){
			int randomCardIndex=rand()%startingPile.size();
			drawPile.push_back(startingPile[randomCardIndex]);
			startingPile.erase(startingPile.begin()+randomCardIndex);
		}

		DrawCard(discardPile,true);

		TwoPlayerButton=new QuickGUI::Button(playerPickMenu,"1 Opponent",{128-42,20},{84,40});
		ThreePlayerButton=new QuickGUI::Button(playerPickMenu,"2 Opponents",{128-42,94},{84,40});
		FourPlayerButton=new QuickGUI::Button(playerPickMenu,"3 Opponents",{128-42,168},{84,40});

		hands.emplace_back();

		for(int i=0;i<7;i++){
			DrawCard(hands[0],true);
		}

		return true;
	}

	Decal*GetCardImage(Color col,CardValue value,bool faceDown=false){
		if(faceDown){
			return IMAGES[IMAGES.size()-1];
		} else
		if(value==WILD||value==DRAWFOUR){
			return IMAGES[4*13+DRAWFOUR-value];
		} else {
			return IMAGES[int(col)*13+value];
		}
	}

	bool OnUserUpdate(float fElapsedTime) override
	{
		switch(state){
			case AIPICK:{ //Player Pick Menu
				playerPickMenu.Update(this);

				bool buttonClicked=false;
				int AICount=0;
				if(TwoPlayerButton->bPressed){
					buttonClicked=true;
					AICount=1;
				}
				if(ThreePlayerButton->bPressed){
					buttonClicked=true;
					AICount=2;
				}
				if(FourPlayerButton->bPressed){
					buttonClicked=true;
					AICount=3;
				}
				if(buttonClicked){
					for(int ai=0;ai<AICount;ai++){
						Hand opponent;
						for (int i=0;i<7;i++){
							DrawCard(opponent);
						}
						hands.push_back(opponent);
					}
					state=GameState::DRAW;
				}
				playerPickMenu.DrawDecal(this);
			}break;
			case DRAW:{
				DrawGameBoard();
				if(!TurnRequiresDraw()){
					state=GameState::PICK;
				}
				if(playerTurn==0){
					DrawStringDecal({0,224},"Player is drawing cards...");	
				} else {
					DrawStringDecal({0,224},"Opponent "+std::to_string(playerTurn)+" is drawing cards...");
				}
				if(GetMouse(0).bPressed&&
					GetMouseX()>=drawPilePos.x&&GetMouseX()<drawPilePos.x+48&&
					GetMouseY()>=drawPilePos.y&&GetMouseY()<drawPilePos.y+48){
					DrawCard(hands[playerTurn],true);
				}
			}break;
			case PICK:{
				DrawGameBoard();
				if(GetMouse(0).bPressed&&
					GetMouseX()>=drawPilePos.x&&GetMouseX()<drawPilePos.x+48&&
					GetMouseY()>=drawPilePos.y&&GetMouseY()<drawPilePos.y+48){
					DrawCard(hands[playerTurn],true);
				}
			}break;
			case RESULT:{
				DrawGameBoard();
			}break;
			case PICKCOLOR:{
				DrawGameBoard();
			}break;
		}
		return true;
	}
};

int main()
{
	CardGameRef demo;
	if (demo.Construct(256, 240, 4, 4))
		demo.Start();
	return 0;
}